############################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
###########################################################################
import os
from M2Crypto import SSL

import ZSI.TC 
from ZSI.TC import _get_global_element_declaration as GED
from ZSI.TC import ElementDeclaration
from ZSI.wstools.Namespaces import WSSE, BEA, GLOBUS, WSTRUST, OASIS

from twisted.python import log
from pyGridWare.utility.gsi import GSIServerContextFactory
from pyGridWare.security.gss import _get_uuid, _is_binary_exchange
from pyGridWare.security.m2.M2SecurityContext import GSIAdapter
from pyGridWare.security.m2.ResourceHome import ManagerHome,\
    Key as GSISecureKey
from pyGridWare.resource.ResourceHome import SimpleResourceHome
import pyGridWare.generated.types.secureconversation
from pyGridWare.generated.services.secureconversation.SecureConversationService import\
    SecureConversationServiceWSRF as WSRFService
    

"""SecureConversation service that uses M2Crypto 
"""

# Typecodes
RequestType = GED(WSTRUST.BASE,"RequestType")
ContextAddress = GED(GLOBUS.SECCONV,"ContextAddress")
SecurityContextToken = GED(BEA.SECCONV,"SecurityContextToken")
TokenType = GED(WSTRUST.BASE,"TokenType")
BinaryExchange = GED(WSTRUST.BASE,"BinaryExchange")

##TODO: As of ZSI-2.0RC2 Union has no pyclass
#  GED RequestType --> GTD RequestTypeOpenEnum -->
#  union of anyURI and RequestTypeEnum
# 
class RequestTypePyclass(str): typecode = RequestType


class SecureConversationService(WSRFService):
    home = ManagerHome.getInstance(SimpleResourceHome)

    def __init__(self, contextFactory):
        """
        Keyword Parameters:
            contextFactory -- security context factory instance 
                eg. pyGridWare.utility.gsi.GSIServerContextFactory()
        """
        self.contextFactory = contextFactory
        WSRFService.__init__(self)

    def wsa_RequestSecurityToken(self, ps, address, **kw):
        log.msg("SecureConversationService.wsa_RequestSecurityToken",
                debug=True)
        request,response = WSRFService.wsa_RequestSecurityToken(self, ps, address, **kw)
        any = filter(_is_binary_exchange, request.Any)
        if len(any) != 1:
            if len(any) < 1: raise RuntimeError, 'Missing BinaryExchange token'
            raise RuntimeError, 'Unsupported multiple BinaryExchange tokens'

        context = GSIAdapter(self.contextFactory.getContext(), connect=False)
        outTokens = context.acceptSecurityContext(any[0])

        key = _get_uuid()
        self.home.add(context, key)
        response.set_attribute_Context('uuid:%s' %key)

        binEx = BinaryExchange.pyclass(outTokens)
        binEx.set_attribute_EncodingType(OASIS.X509TOKEN.Base64Binary)
        binEx.set_attribute_ValueType(GLOBUS.TOKEN)

        ##TODO: WSTRUST.ISSUE is wrong??
        #response.Any = RequestTypePyclass(WSTRUST.ISSUE)
        response.Any = [
            RequestTypePyclass("http://schemas.xmlsoap.org/ws/2004/04/security/trust/Issue"),
            TokenType.pyclass(BEA.SCTOKEN),
            binEx,]
        
        return request,response

    def wsa_RequestSecurityTokenResponse(self, ps, address, **kw):
        log.msg("SecureConversationService.wsa_RequestSecurityTokenResponse",
                debug=True)
        request,response = WSRFService.wsa_RequestSecurityTokenResponse(self, 
                                ps, address, **kw)
        any = filter(_is_binary_exchange, request.Any)
        if len(any) != 1:
            if len(any) < 1: raise RuntimeError, 'Missing BinaryExchange token'
            raise RuntimeError, 'Unsupported multiple BinaryExchange tokens'

        tokens = any[0]
        contextID = request.get_attribute_Context()
        response.set_attribute_Context(contextID)
        
        # remove uuid:
        key = contextID.split(':')[1]
        context = self.home.get(key)
        finish = context.contextCreated()
        outTokens = context.acceptSecurityContext(tokens)
        if finish:
            sct = SecurityContextToken.pyclass()
            response.Any = [
                RequestTypePyclass("http://schemas.xmlsoap.org/ws/2004/04/security/trust/Issue"),
                TokenType.pyclass(BEA.SCTOKEN),
                sct,]

            #TODO: REMOVE OUTDATED
            for what in sct.typecode.ofwhat: what.typed = False

            sct.Identifier = sct.new_Identifier(contextID)
            ctxAddr = ContextAddress.pyclass()
            ctxAddr.Address = ctxAddr.new_Address(address._addressTo)
            ctxAddr.ReferenceProperties = ctxAddr.new_ReferenceProperties()
            ctxAddr.ReferenceProperties.Any = [GSISecureKey(key),]
            sct.Any = [ctxAddr,]
            return request,response

        binEx = BinaryExchange.pyclass(outTokens)
        binEx.set_attribute_EncodingType(OASIS.X509TOKEN.Base64Binary)
        binEx.set_attribute_ValueType(GLOBUS.TOKEN)
        response.Any = [binEx]
        return request,response
    

