#!/usr/bin/env python -W ignore::Warning
##########################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
###########################################################################
import os, sys, shutil, warnings, traceback, re
from optparse import OptionParser
from os.path import isfile, isdir
from ZSI.wstools.WSDLTools import WSDLReader
from ZSI.wstools.XMLSchema import SchemaReader

# set up pyclass metaclass
from ZSI.generate.containers import TypecodeContainerBase, TypesHeaderContainer
TypecodeContainerBase.metaclass = 'pyclass_type'
TypesHeaderContainer.imports.append('from ZSI.generate.pyclass import pyclass_type')

from Modules import WriteServiceModules, \
    WriteClientStubModule, WriteTypesModule, NR
from Utility import CLIENT,SERVICE,\
    SERVICE_STUBS,TYPES,PROPERTIES,RESOURCE


DESC = """
pyGridWare WS-Resource Framework (WSRF) site creation tool.
"""

_get_end_of_path = lambda wsdl: re.sub('[-./:# ]', '_', 
    os.path.splitext(os.path.split(wsdl)[-1])[0].split('?')[0].split('#')[0])

def _debug_cb(option, opt_str, value, parser):
    warnings.resetwarnings()

def _lazy_cb(option, opt_str, value, parser):
    TypecodeContainerBase.lazy = True


def _regenerate_modules(url, name, services=True, stubs=True, xsd=False, **kw):
    '''Regenerate generated modules with new genrator
    '''
    reader = WSDLReader()
    if xsd:
        reader = SchemaReader()

    if isfile(url) is False:
        wsdl = reader.loadFromURL(url)
    else:
        wsdl = reader.loadFromFile(url)

    if xsd:
        class WSDLSchemaWrapper:
            # make xsd processing work by looking like a 
            # <WSDLTools.WSDL >
            def __init__(self, schema, url, name):
                self.types = {schema.targetNamespace:schema}
                self.location = url
                self.name = name

        wsdl = WSDLSchemaWrapper(wsdl, url, name)
        WriteTypesModule(name, wsdl)
        return wsdl

    if services:
        WriteServiceModules(name, wsdl, **kw)
    if stubs:
        WriteClientStubModule(name, wsdl)
    else:
        WriteTypesModule(name, wsdl)
    return wsdl


def _get_config_dir():
    """Find Data Files
    setuptools uses dist.location+'config'
    """
    prefix = sys.prefix
    try:
        import pkg_resources
        d = pkg_resources.get_distribution('pyGridWare')
        if os.path.isdir(os.path.join('/config', d.location)): 
                prefix = d.location
    except:
        pass

    datadir = os.path.join(prefix, 'config')
    return datadir


def main():
    datadir = _get_config_dir()
    sctxt = 'server-config.txt'
    sctac = 'server-config.tac'

    parser = OptionParser(
                 usage="usage: %prog [options] wsdl ...", 
                 description="""pyGridWare WS-Resource Framework (WSRF) site creation tool."""
             )

    # Basic options
    parser.add_option("-f", "--force",
        action="store_true", dest="force", default=False,
        help="do not remove anything, but write over existing generated code.")
        
    parser.add_option("-w", "--web2",
        action="store_true", dest="web2", default=False,
        help="WARNING: unstable. create web2 resource")

    parser.add_option("-l", "--lazy",
        action="callback", callback=_lazy_cb,
        help="WARNING: unstable. lazy evaluation of typecodes")

    parser.add_option("-t", "--top",
        action="store", dest="top", type="string", default=None,
        help="specify top-level module name and others (default to file name). *no multiple args")

    parser.add_option("-r", "--rpy",
        action="store", dest="rpy", default=None, type="string",
        help="basic rpy resource. *no web2, no multiple args")

    parser.add_option("-c", "--script",
        action="store", dest="script", default=None, type="string",
        help="basic client script. *no multiple args")
    
    parser.add_option("-d", "--debug", 
        action="callback", callback=_debug_cb,
        help="debug output")
    
    parser.add_option("-s", "--datadir",
        action="store", dest="datadir", default=datadir, type="string",
        help='data files ("%s", "%s") directory (default "%s")' %(sctxt, sctac, 
                                                                datadir))
        
    (options, args) = parser.parse_args()
    # Test Conditions
    if (len(args) < 1 or (len(args) > 1 and
            not (None is options.script is options.rpy is options.top))): 
        parser.print_help()
        sys.exit(-1)

    run(args[0], config=os.path.join(options.datadir, sctxt),
        tac=os.path.join(options.datadir, sctac),
        top=options.top, 
        force=options.force, 
        web2=options.web2, 
        script=options.script, 
        rpy=options.rpy)



def run(wsdl, config=None, tac=None, top=None, force=False, web2=False, script=None, rpy=None):
    """creates a site
    wsdl -- WSDL file or URL
    Keyword paramters:
         config -- server configuation file, ConfigParser syntax
         tac -- server tac file
         top -- 
         force -- write over 
         web2 -- use twisted.web2 (unstable)
         script -- name of generated client script
         rpy -- name of generated service rpy script
    """
    if script and (not script.endswith('.py')\
            or isfile(script)): 
        raise RuntimeError, 'Client Script name must have suffix ".py" and not exist.'
        sys.exit(-1)
    if rpy and (not rpy.endswith('.rpy')\
            or isfile(os.path.join('services', rpy)) or web2): 
        raise RuntimeError, 'resource file name must have suffix ".rpy" and not exist. no web2.'
        sys.exit(-1)
        
    for i in [config, tac]:
        if not i or not isfile(i):
            raise RuntimeError, "installation missing data file '%s' (try --datadir option)" %i

    # Create README
    if not isfile('README'):
        open('README', 'w').write(README)

    # if not top, remove extension, remove query, remove fragment
    top = top or _get_end_of_path(wsdl)
    wsdl = install(wsdl, mod='generated.%s' %top, smod=top, force=force, web2=web2)
    if not script and not rpy: 
        return
    
    # 
    # WARNING: THIS COULD CHANGE IF ZSI.generate.containers changes.
    #   
    from ZSI.generate.wsdl2dispatch import ServiceModuleWriter
    hack = ServiceModuleWriter()
    try:
        format = {
            'NAME':top,
            'SERVICE':wsdl.services[0].name,
            'LOCATOR':'%sLocator' %wsdl.services[0].name,
            'GETPORTMETHOD':'get%s' %wsdl.services[0].ports[0].name,
            #'RESOURCE':'%sWSRF' %hack.getClassName(wsdl.services[0].name),
            'RESOURCE':'%sWSRF' %wsdl.services[0].name,
        }
        del hack
    except Exception, ex:
        raise

    if script:
        open(script, 'w').write(PYCLIENT %format)
        os.chmod(script, 0755)
    
    if rpy:
        if not isdir('services'): os.mkdir('services')
        open(os.path.join('services', rpy), 'w').write(RPY %format)
    
    format['TXT'] = config
    format['TAC'] = tac
    for src in (config, tac):
        dst = src.rsplit(os.path.sep)[-1]
        if not isfile(dst) or force: shutil.copyfile(src, dst)

    sc = 'start-container.sh'
    if not isfile(sc) or force: 
        open(sc, 'w').write(CONTAINER %format)
        os.chmod(sc, 0755)


def install(wsdl, mod='generated', smod=None, force=False, web2=False, simple=True, 
    stubs=True, services=True, xsd=False):
    """Returns a WSDL instance representing the WSDL.
    wsdl -- WSDL file or URL
    Keyword parameters:
        mod -- containing module for all generated code
        smod -- sub modules for particular WSDL service definition
        force -- write over
        web2 -- TODO fix.  use twisted.web2
    """
    if mod: 
        mod = '%s.' %(mod)
    else: 
        mod = ''

    if not force and isdir(os.path.join(*mod.split('.'))):
        print 'Must remove directory "%s" or specify option force.' %(
            os.path.join(mod.split('.')))
        sys.exit(-1)

    d = dict(module=mod, submodule=smod)
    NR.d = {CLIENT:'%(module)sstubs' %d,
        SERVICE:'%(module)sservices' %d,
        SERVICE_STUBS:'%(module)sservices.stubs' %d,
        TYPES:'%(module)stypes' %d,
        PROPERTIES:'%(module)sproperties' %d,
        RESOURCE:'%(module)sresource' %d,
    }

    name = smod or _get_end_of_path(wsdl)
    try:
        wsdl = _regenerate_modules(wsdl, name=name, services=services, 
                                      stubs=stubs, web2=web2, simple=simple, xsd=xsd)
    except Exception, ex:
        print >>sys.stderr, "Failure: %s" %ex
        traceback.print_exc(sys.stderr)
        sys.exit(-1)

    return wsdl


PYCLIENT = """#!/usr/bin/env python
############################################################################
# Automatically generated by wsdl2web.py
# See LBNLCopyright for copyright notice!
###########################################################################
from twisted.python import log
from twisted.internet import reactor

import ZSI
from pyGridWare.utility.scripts.client import GetBasicOptParser, GetPortKWArgs, SetUp
from generated.%(NAME)s.stubs import %(NAME)s as CLIENT


def main(**kw):
    locator = CLIENT.%(LOCATOR)s()
    port = locator.%(GETPORTMETHOD)s(**kw)
    
    # Factory METHOD Just guessing here
    #response = port.create(CLIENT.CreateRequest())
    #kw['endPointReference'] = response._EndpointReference
    #iport = locator.%(GETPORTMETHOD)s(**kw)
    reactor.stop()


if __name__ == '__main__':
    op = GetBasicOptParser()
    (options, args) = op.parse_args()
    SetUp(options)
    kw = GetPortKWArgs(options)
    reactor.callWhenRunning(main, **kw)
    reactor.run()
"""

RPY = """############################################################################
# Automatically generated by wsdl2web.py
# See LBNLCopyright for copyright notice!
###########################################################################
from generated.%(NAME)s.services.%(NAME)s.%(SERVICE)s import %(RESOURCE)s

class Service(%(RESOURCE)s):
    '''override methods to do something useful'''
resource = Service()
"""

CONTAINER = """#! /bin/sh
############################################################################
# Automatically generated by wsdl2web.py
# See LBNLCopyright for copyright notice!
###########################################################################

EXEC=twistd 
OPTIONS=-noy
CONFIG=server-config.tac

set - ${EXEC} ${OPTIONS} ${CONFIG} "$@"
exec "$@"
"""

README = """############################################################################
# Automatically generated by wsdl2web.py
# See LBNLCopyright for copyright notice!
###########################################################################

Depending on the options provided to "wsdl2web.py", not all the below
files and directories may be present.


FILES: 

    client.py 
      -- Barebones sample client script, sets up a locator and port.

    start-container.sh 
      -- Script for starting the container

    server-config.tac 
      -- Twisted tac file, sets up a few core WSRF services.

    server-config.txt 
      -- Basic server configuration file


DIRECTORIES:

    generated 
      -- Directory with all code generated from the WSDL

    services
      -- Directory to populate with RPY scripts
      -- for "Test.rpy" the path will be "/wsrf/services/Test"


"""
