############################################################################
# Joshua R. Boverhof, LBNL
# See LBNLCopyright for copyright notice!
###########################################################################
import os
from twisted.python import log
from twisted.application.internet import TCPServer
from twisted.protocols.policies import WrappingFactory

from pyGsi import proxylib as _proxylib
from pyGridWare.utility.m2crypto import ContextFactory, GridMapAuthz, SSL
from pyGridWare.utility.m2crypto import TLSProtocolSessionCachingWrapper as TLSProtocolWrapper

"""Must import to do GSI ( I just want to enable proxies here )
Utilities for dealing with GSI and its certificate infrastructure

"""
os.putenv("OPENSSL_ALLOW_PROXY_CERTS", "1")



def GSIServer(port, site, interface=None, contextFactory=None, 
    postConnectionCheck=None):
    """Create a GSI Server, use just like TCPServer or SSLServer
    port -- port number
    site -- twisted.web.server.Site instance
    interface -- optional hostname
    contextFactory -- override default GSI behavior
    postConnectionCheck -- override default GSI behavior
    """
    wrapFactory = wrap_web_site(site, contextFactory=contextFactory, 
                     postConnectionCheck=postConnectionCheck) 

    server = TCPServer(port, wrapFactory, interface=interface)

    # Not really good way to get at this.
    server.contextFactory = contextFactory
    return server


def wrap_web_site(site, contextFactory=None, postConnectionCheck=None):
    """setup GSI twisted.web.server.Site
    root -- Resource Tree
    contextFactory -- 
    postConnectionCheck -- 
    """
    contextFactory= contextFactory or GSIServerContextFactory()
    postConnectionCheck = postConnectionCheck or GridMapAuthz()

    site.startTLS = True
    site.sslChecker = SSL.Checker.Checker()

    wrapFactory = WrappingFactory(site)
    wrapFactory.protocol.TLS = True
    wrapFactory.protocol = lambda factory, wrappedProtocol: \
        TLSProtocolWrapper(factory,
                           wrappedProtocol,
                           startPassThrough=0,
                           client=0,
                           contextFactory=contextFactory,
                           postConnectionCheck=postConnectionCheck)
    return wrapFactory



def get_user_credentials(cert=None, key=None, calist=()):
    """Use cert and key, else proxy certificate, else default "usercert"
    Add GSI CA certifcates to calist
    """
    userdir = os.path.expanduser('~')
    if not cert:
        try:
            cert = key = _proxylib.get_proxy_filename()
            proxy = _proxylib.get_usercert(cert)
        except:
            cert = os.path.join(userdir, '.globus', 'usercert.pem')
            key = os.path.join(userdir, '.globus', 'userkey.pem')
            if not os.path.isfile(cert):
                raise RuntimeError, 'GSI: no certificate "%s"' %cert

            if key and not os.path.isfile(key):
                raise RuntimeError, 'GSI: no key "%s"' %key

        else:
            if not _proxylib.check_valid(proxy):
                raise _proxylib.InvalidProxyException

    calist = list(calist)
    calist += get_default_calist()

    return cert,key,tuple(calist)


def get_default_calist():
    userdir = os.path.expanduser('~')
    capath = os.path.join(userdir, '.globus', 'certificates')
    if not os.path.isdir(capath):
        capath = os.path.join(os.sep, 'etc', 'grid-security', 'certificates')
        if not os.path.isdir(capath):
            log.debug('GSI: no trusted CA certificates "%s"' %certificates, debug=True)
            return []

    return get_calist(capath)
       
def get_calist(capath):
    return [os.path.join(capath,i) for i in filter(lambda f: f.endswith('.0'), 
        os.listdir(capath))]


class GSIServerContextFactory(ContextFactory):
    """Copied from twisted.internet.ssl.DefaultOpenSSLContextFactory,
    which can not be imported.

    Class Attributes:
        containercert -- default GSI container certificate
        containerkey --  default GSI container key

    """
    containercert = os.path.join(os.sep, 'etc', 'grid-security', 'containercert.pem')
    containerkey = os.path.join(os.sep, 'etc', 'grid-security', 'containerkey.pem')
    
    def __init__(self, privateKeyFileName=None, certificateFileName=None,
                 sslmethod=SSL.SSLv3_METHOD, calist=(), capath=''):
        """
        @param privateKeyFileName: Name of a file containing a private key
        @param certificateFileName: Name of a file containing a certificate
        @param sslmethod: The SSL method to use
        """
        self.privateKeyFileName = privateKeyFileName or self.containerkey
        self.certificateFileName = certificateFileName or self.containercert
        self.sslmethod = sslmethod
        #self.calist = list(calist) + get_default_calist()
        self.calist = list(calist)
        # not sure about the ramifications of specifying capath
        self.capath = capath
        self.cacheContext()

    def cacheContext(self):
        ctx = SSL.Context(self.sslmethod)
        ctx.set_options(SSL.SSL_OP_DONT_INSERT_EMPTY_FRAGMENTS)
        ctx.set_cipher_list("NULL-MD5:ALL:!ADH:!EXP:@STRENGTH")
        ctx.set_allow_unknown_ca(False)
        ctx.set_verify(SSL.verify_peer, 10)

        ctx.load_cert_chain(self.certificateFileName,
                             keyfile=self.privateKeyFileName)
        for ca in self.calist: 
            ctx.load_verify_locations(cafile=ca, capath=self.capath)

        self._context = ctx

    def __getstate__(self):
        d = self.__dict__.copy()
        del d['_context']
        return d

    def __setstate__(self, state):
        self.__dict__ = state
        self.cacheContext()

    def getContext(self):
        """Create an SSL context.
        """
        return self._context


class GSIClientContextFactory(ContextFactory):
    """Copied from twisted.internet.ssl.DefaultOpenSSLContextFactory,
    which can not be imported."""
    
    def __init__(self, privateKeyFileName=None, certificateFileName=None,
                 sslmethod=SSL.SSLv3_METHOD, calist=(), capath=''):
        """
        @param privateKeyFileName: Name of a file containing a private key
        @param certificateFileName: Name of a file containing a certificate
        @param sslmethod: The SSL method to use
        """
        self.sslmethod = sslmethod
        self.capath = capath
        self.privateKeyFileName,self.certificateFileName,self.calist = \
            get_user_credentials(key=privateKeyFileName, 
                                cert=certificateFileName, 
                                calist=calist)
        self.cacheContext()

    def cacheContext(self):
        ctx = SSL.Context(self.sslmethod)
        ctx.set_options(SSL.SSL_OP_DONT_INSERT_EMPTY_FRAGMENTS)

        log.msg('client cred: ', self.certificateFileName, self.privateKeyFileName, debug=True)
        ctx.load_cert_chain(self.certificateFileName, keyfile=self.privateKeyFileName)

        log.msg('CA LIST: ', self.calist, debug=True)
        for ca in self.calist: 
            ctx.load_verify_locations(cafile=ca, capath=self.capath)

        self._context = ctx

    def __getstate__(self):
        d = self.__dict__.copy()
        del d['_context']
        return d

    def __setstate__(self, state):
        self.__dict__ = state
        self.cacheContext()

    def getContext(self):
        """Create an SSL context.
        """
        return self._context



